@@ -11,7 +11,7 @@ wget_cmd = lambda h : ["wget", "-O", "/dev/null", "https://" + h]
 for fname in sys.argv[1:]:
   hosts = host_targets.findall(open(fname).read())
   if not hosts:
-    print "Could not find <target hosts> in " + fname
+    print("Could not find <target hosts> in " + fname)
     continue
   successes = []
   failures = []
@@ -26,15 +26,15 @@ for fname in sys.argv[1:]:
     else:
       successes.append(h)
   if successes and not failures:
-    print fname + " no cert warnings"
+    print(fname + " no cert warnings")
   elif failures and not successes:
-    print fname + " categorical failure:"
+    print(fname + " categorical failure:")
     for f in failures:
-      print "    " + f
+      print("    " + f)
   else:
-    print fname + " mixed results:"
+    print(fname + " mixed results:")
     for s in successes:
-      print "    " + s + " is OK"
+      print("    " + s + " is OK")
     for f in failures:
-      print "    " + f
+      print("    " + f)
 
